use std::{io::{Read, Cursor}, fmt::Display};
use anyhow::Result;

use crate::packets::{Decode, Encode};

#[derive(Debug, PartialEq, Eq)]
pub struct VarInt {
    pub value: i32,
}

impl Decode for VarInt {
    fn decode(cursor: &mut Cursor<&[u8]>) -> Result<VarInt> {
        let mut value: i32 = 0;
        let mut pos = 0;
        let mut current_byte = vec![0u8; 1];

        loop {
            cursor.read_exact(&mut current_byte)?;

            value |= (current_byte[0] as i32 & 0x7F) << pos;
            if current_byte[0] & 0x80 == 0 {
                break;
            }
            pos += 7;
        }

        Ok(VarInt { value })
    }
}

impl Encode for VarInt {
    fn encode(&self, buffer: &mut Vec<u8>) -> Result<()> {
        let mut value = self.value as u32;

        loop {
            if value & !0x7F == 0 {
                buffer.push(value as u8);
                break;
            }

            buffer.push(((value & 0x7F) | 0x80) as u8);

            value >>= 7;
        }

        Ok(())
    }
}


impl From<VarInt> for i32 {
    fn from(val: VarInt) -> Self {
        val.value
    }
}

impl Display for VarInt {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.value.fmt(f)
    }
}

#[cfg(test)]
mod test {
    use std::io::Cursor;

    use crate::datatypes::varint::VarInt;
    use crate::packets::{Decode, Encode};

    #[test]
    fn small_number_decode_works() {
        let data = vec![0x10];
        let mut cursor = Cursor::new(&data[..]);
        let result = VarInt::decode(&mut cursor).unwrap();
        
        assert_eq!(16, result.value);
        assert_eq!(1, cursor.position());
    }

    #[test]
    fn large_number_decode_works() {
        let data = vec![0xff, 0xff, 0x7f];
        let mut cursor = Cursor::new(&data[..]);
        let result = VarInt::decode(&mut cursor).unwrap();
        
        assert_eq!(2097151, result.value);
        assert_eq!(3, cursor.position());
    }

    #[test]
    fn negative_number_decode_works() {
        let data = vec![0xff, 0xff, 0xff, 0xff, 0x0f];
        let mut cursor = Cursor::new(&data[..]);
        let result = VarInt::decode(&mut cursor).unwrap();
        
        assert_eq!(-1, result.value);
        assert_eq!(5, cursor.position());
    }

    #[test]
    fn small_number_encode_works() {
        let mut data = Vec::new();
        let to_encode = VarInt {value: 12};

        to_encode.encode(&mut data);

        assert_eq!(vec![12], data);
    }


    #[test]
    fn large_number_encode_works() {
        let mut data = Vec::new();
        let to_encode = VarInt {value: 2097151};

        to_encode.encode(&mut data);

        assert_eq!(vec![0xff, 0xff, 0x7f], data);
    }

    #[test]
    fn negative_number_encode_works() {
        let mut data = Vec::new();
        let to_encode = VarInt {value: -1};

        to_encode.encode(&mut data);

        assert_eq!(vec![0xff, 0xff, 0xff, 0xff, 0x0f], data);
    }
}
